cimport cython
cimport cython.array
from libc.stdlib cimport malloc, free

@cython.boundscheck(False)
def decompress(string, usize):
    if not string:
        return ''
    cdef unsigned data_pos = 0, datalen = len(string)
    cdef unsigned char* data = string
    cdef unsigned int out_pos = 0, outlen = usize
    cdef unsigned char* output = <unsigned char*> malloc(outlen * sizeof(unsigned char))
    cdef unsigned flags = 0, flag_mask = 0
    cdef unsigned ref_in = 0, ref_len = 0

    try:
        while data_pos < datalen:
            if not flag_mask:
                if data_pos + 1 >= datalen:
                    raise RuntimeError('decompression failure: insufficient data')
                flags = data[data_pos]
                data_pos += 1
                flag_mask = 0x80

            if flags & flag_mask:
                if data_pos + 2 >= datalen:
                    raise RuntimeError('decompression failure: insufficient data')
                ref_in = data[data_pos]
                data_pos += 1
                ref_len = data[data_pos]
                data_pos += 1

                ref_in |= (ref_len & 0x0F) << 8
                ref_in = out_pos - ref_in
                ref_len = ((ref_len & 0xF0) >> 4) + 1
                if ref_in < 0:
                    raise RuntimeError('decompression failure: insufficient buffer')

                if out_pos+ref_len > outlen:
                    raise RuntimeError('decompression failure: excessive data', out_pos, outlen)
                for i in xrange(ref_len):
                    output[out_pos] = output[ref_in]
                    ref_in += 1
                    out_pos += 1

            else:
                if out_pos >= outlen:
                    raise RuntimeError('decompression failure: excessive data', data_pos, out_pos, outlen)
                output[out_pos] = data[data_pos]
                out_pos += 1
                data_pos += 1

            flag_mask >>= 1

        string = <bytes> output[:out_pos]
    finally:
        free(output)

    return string

def parsebitmap(stream, context):
    cdef unsigned w = context.w
    cdef unsigned h = context.h
    cdef unsigned bmpw = ((w - 1) / 4 + 1) * 4
    cdef unsigned srcw = w / 4 * 4
    cdef unsigned char* output = <unsigned char*> malloc(bmpw * h * sizeof(unsigned char))
    cdef unsigned x = 0, y = 0, repeat = 0
    cdef unsigned cur, last, havelast = 0
    cdef unsigned char* buf
    cdef unsigned i, j, m

    try:
        while x < srcw:
            s = stream.read(4)
            cur = (<unsigned*><char*> s)[0]
            
            if not repeat:
                repeat = 1
            elif havelast and cur == last:
                repeat += 1
            
            j = bmpw*(h-y-1)+x
            (<unsigned*> output)[j/4] = cur
            y += 1
            
            if y == h: 
                y = 0
                x += 4
                if repeat == 1:
                    repeat = 0
            
            if repeat:
                last = cur
                havelast = 1
            
            if repeat == 2:
                s = stream.read(1)
                m = (<unsigned char*>s)[0]
                if not m:
                    s = stream.read(1)
                    m = 256 + (<unsigned char*>s)[0]
                
                until = m + y - 1
                while y < until:
                    j = bmpw*(h-y-1)+x
                    (<unsigned*> output)[j/4] = cur
                    y += 1
                if y == h:
                    y = 0
                    x += 4
                repeat = 0
    
        string = <bytes> output[:bmpw * h]
    finally:
        free(output)

    return string