# encoding=utf_8
import sys
import os
import csv
import struct
import zipfile
import time
import math
import wave
import ctypes
from collections import defaultdict, deque
from cStringIO import StringIO
from ConfigParser import RawConfigParser
from traceback import print_exc, print_stack
from threading import Lock
from random import randint
from ctypes import c_long, c_int
import wx.grid as wxgrid

import wx
import wx.lib.delayedresult as delayedresult
import wx.lib.newevent
import pyaudio

from sogna.dat import DATArchive
from sogna.win import bytecode
from sogna.interpreter import *
from sogna.anm import Animation

#   TODO (low-priority):
# i18n operators
# fullscreen
#   BUGS
# some fade ins, the screen shows up before the fade even starts.
#  oddly, the bytecode confirms that's what should be happening

if sys.platform == 'win32':
    SetWindowLong = ctypes.windll.user32.SetWindowLongW
    CallWindowProc = ctypes.windll.user32.CallWindowProcW
    WndProcType = ctypes.WINFUNCTYPE(c_int, c_long, c_int, c_int, c_int)
    GWL_WNDPROC = -4
    MM_MCINOTIFY = 0x3B9

    winmm = ctypes.windll.winmm
    mcibuf = ctypes.create_string_buffer(1024)
    def mciSendString(s, buf=mcibuf, len=1024, hwnd=0):
        i = winmm.mciSendStringA(s, buf, len, hwnd)
        if i != 0:
            raise RuntimeError("Error in mciSendString", i, s) # TODO proper ctypes handler
        return buf.value
else: # TODO multi-platform MIDI
    def mciSendString(*args, **kwargs):
        return

NextStateEvent, EVT_NEXT_STATE = wx.lib.newevent.NewEvent()
SoundFinishedEvent, EVT_SOUND_FINISHED = wx.lib.newevent.NewEvent()

DEFAULT_CAPTION = 'PySGS'

class MaskDict(dict):
    def __init__(self, nparr):
        self.nparr = nparr
    def __missing__(self, key):
        mask = self.nparr != key
        self[key] = mask
        return mask

def withmask(image):
    if not hasattr(image, 'nparray'):
        image.nparray = np.frombuffer(image.data, dtype=np.uint8).reshape(image.h, len(image.data) / image.h)[::-1, :image.w]
        image.masks = MaskDict(image.nparray)
    return image

class MixerDialog(wx.Dialog):
    def __init__(self, game, *args, **kwargs):
        super(MixerDialog, self).__init__(*args, **kwargs)
        self.game = game
        self.origvoicevol = game.states.voice_volume
        self.origmusicvol = game.states.music_volume
        voicevol = self.origvoicevol >> 8
        musicvol = self.origmusicvol >> 8

        self.panel = wx.Panel(self)
        sizer = wx.GridBagSizer(5, 5)
        sizer.AddGrowableCol(1)

        self.voiceslider = wx.Slider(self.panel, value=voicevol, maxValue=256)
        self.voiceslider.SetMinSize((300, -1))
        sizer.Add(wx.StaticText(self.panel, label='Voice'), (0,0))
        sizer.Add(self.voiceslider, (0, 1), flag=wx.EXPAND)
        self.Bind(wx.EVT_SCROLL, lambda e: self.SetVoiceVolume(e.Selection << 8), self.voiceslider)
        self.musicslider = wx.Slider(self.panel, value=musicvol, maxValue=256)
        sizer.Add(wx.StaticText(self.panel, label='Music'), (1,0))
        sizer.Add(self.musicslider, (1, 1), flag=wx.EXPAND)
        self.Bind(wx.EVT_SCROLL, lambda e: self.SetMusicVolume(e.Selection << 8), self.musicslider)

        bsizer = wx.StdDialogButtonSizer()
        self.btn_close = wx.Button(self.panel, wx.ID_OK)
        bsizer.AddButton(self.btn_close)
        self.btn_cancel = wx.Button(self.panel, wx.ID_CANCEL)
        bsizer.AddButton(self.btn_cancel)
        bsizer.Realize()
        sizer.Add(bsizer, (2, 0), (1, 2), flag=wx.EXPAND)
        self.Bind(wx.EVT_BUTTON, self.OnCloseButton, self.btn_close)
        self.Bind(wx.EVT_BUTTON, self.OnCancelButton, self.btn_cancel)

        self.panel.SetSizer(sizer)
        dsizer = wx.BoxSizer(wx.VERTICAL)
        dsizer.Add(self.panel, 1, wx.EXPAND | wx.ALL, 5)
        self.SetSizerAndFit(dsizer)

    def OnCloseButton(self, event):
        self.Show(False)

    def OnCancelButton(self, event):
        self.Show(False)
        self.SetVoiceVolume(self.origvoicevol)
        self.SetMusicVolume(self.origmusicvol)

    def SetVoiceVolume(self, value):
        gm = self.game
        st = gm.states
        newvol = min(value, 0xFFFF)
        if newvol != st.voice_volume:
            st.voice_volume = newvol
            gm.setvolumes()

    def SetMusicVolume(self, value):
        gm = self.game
        st = gm.states
        newvol = min(value, 0xFFFF)
        if newvol != st.music_volume:
            st.music_volume = newvol
            gm.setvolumes()

class JumpPanel(wx.Panel):
    def __init__(self, parent, game):
        super(JumpPanel, self).__init__(parent)
        self.game = game
        self.winnames = []
        self.ops = []

        splitter = wx.SplitterWindow(self)
        splitter.SetMinimumPaneSize(100)
        self.winlist = wx.ListBox(splitter, style=wx.LB_SINGLE)
        self.Bind(wx.EVT_LISTBOX, self.OnWinSelect, self.winlist)
        self.opslist = wx.ListCtrl(splitter, style=wx.LC_SINGLE_SEL|wx.LC_REPORT)
        self.opslist.InsertColumn(0, 'Addr', wx.LIST_FORMAT_RIGHT, 50)
        self.opslist.InsertColumn(1, 'Op', wx.LIST_FORMAT_RIGHT, 40)
        self.opslist.InsertColumn(2, 'Mnem')
        self.opslist.InsertColumn(3, 'Values', width=400)
        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.OnOpSelect, self.opslist)

        splitter.SplitVertically(self.winlist, self.opslist)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(splitter, 1, flag=wx.EXPAND)
        self.SetSizer(sizer)

        self.gamewin = None

    def SetWinNames(self, names):
        self.winnames = names
        self.winlist.SetItems(names)
        self.ops = []
        self.RefreshOps()

    def setcurrentwin(self, name, ops):
        self.gamewin = name+'.WIN'
        if self.winlist.GetStringSelection() != self.gamewin:
            self.winlist.SetStringSelection(self.gamewin)
            self.ops = ops
            self.RefreshOps()

    def setopindex(self, index):
        if self.winlist.GetStringSelection() == self.gamewin:
            ctrl = self.opslist
            if ctrl.GetFirstSelected() != index:
                perpage = ctrl.GetCountPerPage()
                ctrl.EnsureVisible(index - perpage / 3)
                ctrl.EnsureVisible(index + perpage / 3)
                ctrl.Select(index)

    def OnWinSelect(self, event):
        self.ops = self.game.loadwin(self.winnames[event.Selection])
        self.RefreshOps()

    def RefreshOps(self):
        lst = self.opslist
        lst.DeleteAllItems()
        for op in self.ops:
            lst.Append(['%06X'%op.address, '%02X'%op.instruction, op.fourcc, joinvalues(op)])

    def OnOpSelect(self, event):
        game = self.game
        if not game.IsGameRunning():
            st = game.states
            gowin = self.winlist.GetStringSelection()
            if gowin != self.gamewin:
                game.LoadWin(gowin)
            game.states.index = event.Index

def joinvalues(op):
    parts = []
    for n, v in op:
        if n in ('address', 'instruction', 'fourcc', 'nextaddr'):
            continue
        if isinstance(v, (Container, list)):
            v = '...'
        parts.append('%s=%s'%(n, v))
    return ' '.join(parts)

class GlobalsTable(wxgrid.PyGridTableBase):
    def __init__(self):
        super(GlobalsTable, self).__init__()
        self.data = [0] * 256
    #--------------------------------------------------
    # required methods for the wxPyGridTableBase interface
    def GetNumberRows(self):
        return 16
    def GetNumberCols(self):
        return 16
    def IsEmptyCell(self, row, col):
        try:
            return not self.data[16*row+col]
        except IndexError:
            return True
    def GetValue(self, row, col):
        try:
            return self.data[16*row+col]
        except IndexError:
            return ''
    def SetValue(self, row, col, value):
        try:
            self.data[16*row+col] = value
        except IndexError:
            return
    #--------------------------------------------------
    def GetColLabelValue(self, col):
        return '_%X'%col
    def GetRowLabelValue(self, row):
        return '%X_'%row
    def GetTypeName(self, row, col):
        return wxgrid.GRID_VALUE_NUMBER
    # Called to determine how the data can be fetched and stored by the
    # editor and renderer.  This allows you to enforce some type-safety
    # in the grid.
    def CanGetValueAs(self, row, col, typeName):
        if typeName == wxgrid.GRID_VALUE_NUMBER:
            return True
        else:
            return False
    def CanSetValueAs(self, row, col, typeName):
        return self.CanGetValueAs(row, col, typeName)

class GlobalsPanel(wx.Panel):
    def __init__(self, *args, **kwargs):
        super(GlobalsPanel, self).__init__(*args, **kwargs)
        self.grid = grid = wxgrid.Grid(self)
        grid.CreateGrid(16, 16)
        self.table = GlobalsTable()
        grid.SetDefaultColSize(40, True)
        grid.SetTable(self.table, True)
        grid.SetRowLabelSize(grid.GetDefaultColSize())
        grid.SetColLabelSize(grid.GetDefaultRowSize())
        grid.SetMargins(0,0)
        grid.EnableEditing(False)

        sizer = wx.GridBagSizer(5, 5)
        sizer.Add(self.grid, (0,0))
        self.SetSizer(sizer)

    def SetDataSource(self, data):
        self.table.data = data
        self.UpdateData()

    def UpdateData(self):
        msg = wxgrid.GridTableMessage(self.table, wxgrid.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
        self.table.GetView().ProcessTableMessage(msg)

class DebugFrame(wx.Frame):
    def __init__(self, game, *args, **kwargs):
        super(DebugFrame, self).__init__(*args, **kwargs)
        self.Bind(wx.EVT_CLOSE, self.OnClose)
        self.panel = wx.Panel(self)

        self.game = game

        self.fps = wx.TextCtrl(self.panel, style=wx.TE_READONLY)

        self.pausebtn = wx.ToggleButton(self.panel, label='&Pause')
        self.Bind(wx.EVT_TOGGLEBUTTON, self.OnPauseButton, self.pausebtn)
        self.loadbutton = wx.Button(self.panel, label='&Load')
        self.loadbutton.Enable(False)
        self.Bind(wx.EVT_BUTTON, self.OnLoadButton, self.loadbutton)
        self.searchbutton = wx.Button(self.panel, label='&Search')
        self.Bind(wx.EVT_BUTTON, self.OnSearchButton, self.searchbutton)
        self.nextbutton = wx.Button(self.panel, label='Next')
        self.Bind(wx.EVT_BUTTON, self.OnNextButton, self.nextbutton)
        self.forbutton = wx.Button(self.panel, label='Forward')
        self.Bind(wx.EVT_BUTTON, self.OnForwardButton, self.forbutton)

        self.dialogtext = wx.TextCtrl(self.panel, style=wx.TE_READONLY|wx.TE_MULTILINE)

        self.jumppanel = JumpPanel(self.panel, game)
        jumpsizer = wx.BoxSizer(wx.VERTICAL)
        jumpsizer.Add(wx.StaticText(self.panel, label='Bytecode'), flag=wx.LEFT, border=5)
        jumpsizer.Add(self.jumppanel, 1, wx.EXPAND)
        self.SetWinNames = self.jumppanel.SetWinNames
        self.setcurrentwin = self.jumppanel.setcurrentwin
        self.jumppanel.Enable(False)

        self.globalspanel = GlobalsPanel(self.panel)
        globalsizer = wx.BoxSizer(wx.VERTICAL)
        globalsizer.Add(wx.StaticText(self.panel, label='Globals'), flag=wx.LEFT, border=5)
        globalsizer.Add(self.globalspanel, 1, wx.EXPAND)
        self.globalspanel.SetMinSize(self.globalspanel.GetBestSize())
        self.globalspanel.Enable(False)

        self.inslist = wx.TextCtrl(self.panel, value='52')
        self.varlist = wx.TextCtrl(self.panel, value='16 17')

        self.opslist = wx.ListCtrl(self.panel, style=wx.LC_SINGLE_SEL|wx.LC_REPORT)
        self.opslist.InsertColumn(0, 'WIN')
        self.opslist.InsertColumn(1, 'Addr', wx.LIST_FORMAT_RIGHT, 50)
        self.opslist.InsertColumn(2, 'Op', wx.LIST_FORMAT_RIGHT, 40)
        self.opslist.InsertColumn(3, 'Mnem')
        self.opslist.InsertColumn(4, 'Values', width=400)
        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.OnOpSelect, self.opslist)

        def boxsizer(sflags, wflags, widgets):
            sizer = wx.BoxSizer(sflags)
            for widget in widgets:
                sizer.Add(widget, flag=wflags)
            return sizer

        stflags = wx.LEFT|wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL
        sizer = wx.GridBagSizer(5, 5)
        sizer.AddGrowableRow(3, 1)
        sizer.AddGrowableRow(6, 1)
        sizer.AddGrowableCol(1)
        sizer.Add(wx.StaticText(self.panel, label='FPS'), (1, 0), flag=stflags, border=5)
        sizer.Add(self.fps, (1, 1))
        sizer.Add(wx.StaticText(self.panel, label='Dialogue'), (2, 0), flag=stflags, border=5)
        sizer.Add(self.dialogtext, (2, 1), (1, 3), wx.EXPAND)
        sizer.Add(boxsizer(wx.HORIZONTAL, wx.DEFAULT, [self.pausebtn, self.loadbutton, self.searchbutton, self.nextbutton, self.forbutton]), (0, 0), (1, 3))
        sizer.Add(jumpsizer, (3, 0), (1, 3), wx.EXPAND)
        sizer.Add(globalsizer, (4, 0), (1, 3), wx.EXPAND)
        sizer.Add(boxsizer(wx.HORIZONTAL, wx.DEFAULT, [self.inslist, self.varlist]), (5, 0), flag=wx.EXPAND)
        sizer.Add(self.opslist, (6, 0), (1, 3), wx.EXPAND)
        self.panel.SetSizer(sizer)

    def nostate(self, states):
        self.nextbutton.Enable(False)
        self.nextbutton.SetLabel('none')

    def nextstate(self, states, state):
        self.nextbutton.SetLabel('next (%d)'%states.state)
        self.nextbutton.Enable(True)
        self.jumppanel.setopindex(states.index)
        self.globalspanel.UpdateData()

    def nextop(self, states, op):
        pass

    def OnClose(self, event):
        self.game.debugitem.Check(False)
        self.Show(False)
        event.Veto(True)

    def OnNextButton(self, event):
        self.nextbutton.Enable(False)
        self.nextbutton.SetLabel('next')
        states = self.game.states
        self.game.clearstate()
        states.pause_anim = states.future_pause
        states.future_pause = 0

    def OnPauseButton(self, event):
        if event.Selection:
            self.game.StopGame()
            self.jumppanel.Enable(True)
            self.globalspanel.Enable(True)
            self.loadbutton.Enable(self.game.states.loadhandler is not None)
        else:
            self.game.StartGame()
            self.jumppanel.Enable(False)
            self.globalspanel.Enable(False)
            self.loadbutton.Enable(False)
        self.globalspanel.grid.EnableEditing(event.Selection)

    def OnForwardButton(self, event):
        self.game.OnAnimationTimer(event)

    def OnLoadButton(self, event):
        states = self.game.states
        states.setaddress(states.loadhandler)
        states.resetandnext()
        self.game.StartGame()
        self.jumppanel.Enable(False)
        self.globalspanel.Enable(False)
        self.loadbutton.Enable(False)
        self.globalspanel.grid.EnableEditing(False)
        self.pausebtn.SetValue(False)

    def OnSearchButton(self, event):
        matched = []
        codes = map(int, self.inslist.GetValue().split())
        vars = map(int, self.varlist.GetValue().split())
        for name in self.jumppanel.winnames:
            name = name[:-4]
            ops = self.game.loadwin(name)
            for op in ops:
                if op.instruction in codes and op.a in vars:
                    matched.append((name, op))
        lst = self.opslist
        lst.DeleteAllItems()
        self.matchedops = matched
        for win, op in matched:
            lst.Append([win, '%06X'%op.address, '%02X'%op.instruction, op.fourcc, joinvalues(op)])

    def OnOpSelect(self, event):
        st = self.game.states
        win, op = self.matchedops[event.Index]
        st.functions[op.instruction](op)
        self.globalspanel.UpdateData()
        self.OnLoadButton(None)

class Sound(object):
    __slots__ = ['data', 'nextdata', 'index', 'vol1', 'vol2', 'loop', 'channels', 'format']
    def __init__(self):
        self.data = None # 0, 2
        self.nextdata = None # 3
        self.index = 0
        self.vol1 = 256 # 4
        self.vol2 = 256 # 5
        self.loop = 0 # 6
        self.channels = 0
        self.format = None
    def getdata(self):
        data, fmt = self.data, self.format
        if data is not None:
            if fmt == pyaudio.paInt16:
                data = data.view(np.int16)
            elif fmt == pyaudio.paInt8:
                data = data.view(np.int8)
            else:
                raise NotImplementedError('unknown audio format', fmt)
        return self.index, data
    def loopornext(self):
        self.index = 0
        if not self.loop:
            data = self.nextdata
            if data is None or not len(data):
                self.data = None
            else:
                self.data, self.channels, self.format = data
        return self.getdata()

class GameFrame(wx.Frame):
    def __init__(self, parent):
        style = wx.DEFAULT_FRAME_STYLE & ~(wx.MAXIMIZE_BOX | wx.RESIZE_BORDER)
        super(GameFrame, self).__init__(parent, title=DEFAULT_CAPTION, style=style)
        self.Bind(wx.EVT_CLOSE, self.OnClose)
        if sys.platform == 'win32':
            self.newWndProc = WndProcType(self.MyWndProc)
            self.oldWndProc = SetWindowLong(self.GetHandle(), GWL_WNDPROC, self.newWndProc )

        self.debug = DebugFrame(self, self, size=(500, 500))

        self.menubar = wx.MenuBar()
        self.SetMenuBar(self.menubar)

        self.savemenu = wx.Menu()
        for i in xrange(10):
            saveitem = self.savemenu.Append(wx.ID_ANY, 'Data%d'%i)
            self.Bind(wx.EVT_MENU, lambda e, n=i: self.SaveSaveData(n), saveitem)
        self.menubar.Append(self.savemenu,"&Save")
        self.menubar.EnableTop(0, False)

        self.loadmenu = wx.Menu()
        self.loaditems = []
        for i in xrange(10):
            loaditem = self.loadmenu.Append(wx.ID_ANY, 'Data%d'%i)
            loaditem.Enable(False)
            self.Bind(wx.EVT_MENU, lambda e, n=i: self.LoadSaveData(n), loaditem)
            self.loaditems.append(loaditem)
        self.menubar.Append(self.loadmenu,"&Load")
        self.menubar.EnableTop(1, False)

        self.musicmenu = wx.Menu()
        self.volumeitem = self.musicmenu.Append(wx.ID_ANY, '&Volume')
        self.Bind(wx.EVT_MENU, self.OnUpdateVolumes, self.volumeitem)
        self.musicmenu.AppendSeparator()
        self.musicitem = self.musicmenu.AppendCheckItem(wx.ID_ANY, 'Music')
        self.musicitem.Check()
        self.Bind(wx.EVT_MENU, self.OnConfigControlUpdated, self.musicitem)
        self.voiceitem = self.musicmenu.AppendCheckItem(wx.ID_ANY, 'Voice')
        self.voiceitem.Check()
        self.Bind(wx.EVT_MENU, self.OnConfigControlUpdated, self.voiceitem)
        self.musicmenu.AppendSeparator()
        self.pcmaudio8 = self.musicmenu.AppendRadioItem(wx.ID_ANY, '8-bit')
        self.pcmaudio8.Check()
        self.pcmaudio16 = self.musicmenu.AppendRadioItem(wx.ID_ANY, '16-bit')
        self.Bind(wx.EVT_MENU, self.OnConfigControlUpdated, self.pcmaudio16)
        self.Bind(wx.EVT_MENU, self.OnConfigControlUpdated, self.pcmaudio8)
        self.pcmimmediate = self.musicmenu.AppendCheckItem(wx.ID_ANY, 'Play PCM Immediately')
        self.Bind(wx.EVT_MENU, self.OnConfigControlUpdated, self.pcmimmediate)
        self.menubar.Append(self.musicmenu,"&Audio")

        self.exitmenu = wx.Menu()
        self.exititem = self.exitmenu.Append(wx.ID_EXIT, 'E&xit Windows')
        self.Bind(wx.EVT_MENU, self.OnExit, self.exititem)
        self.restartitem = self.exitmenu.Append(wx.ID_ANY, 'Exi&t Menu')
        self.restartitem.Enable(False)
        self.Bind(wx.EVT_MENU, self.OnRestart, self.restartitem)
        self.menubar.Append(self.exitmenu,"&Exit")

        self.othermenu = wx.Menu()
        self.repeatitem = self.othermenu.Append(wx.ID_ANY, 'Repeat')
        self.repeatitem.Enable(False)
        self.Bind(wx.EVT_MENU, self.OnRepeat, self.repeatitem)
        self.debugitem = self.othermenu.AppendCheckItem(wx.ID_ANY, 'Debug')
        self.Bind(wx.EVT_MENU, lambda e: self.debug.Show(self.debugitem.IsChecked()), self.debugitem)
        self.menubar.Append(self.othermenu,"Other")

        self.bmppanel = wx.Panel(self, style=wx.WANTS_CHARS|wx.NO_BORDER)
        self.bmppanel.SetBackgroundColour('BLACK')
        self.bmppanel.Bind(wx.EVT_LEFT_DOWN, self.OnMouseButton)
        self.bmppanel.Bind(wx.EVT_LEFT_UP, self.OnMouseButton)
        self.bmppanel.Bind(wx.EVT_RIGHT_DOWN, self.OnMouseButton)
        self.bmppanel.Bind(wx.EVT_RIGHT_UP, self.OnMouseButton)
        self.bmppanel.Bind(wx.EVT_MIDDLE_DOWN, self.OnMouseButton)
        self.bmppanel.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.bmppanel.Bind(wx.EVT_KEY_UP, self.OnKeyUp)
        self.bmppanel.Bind(wx.EVT_MOTION, self.OnMouseMoved)
        self.bmppanel.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.Bind(wx.EVT_KEY_UP, self.OnKeyUp)

        self.dialoguepopup = wx.Dialog(self, style=0)
        self.Bind(wx.EVT_MOVE, self.OnMove)
        self.dialoguepopup.Bind(wx.EVT_CLOSE, lambda e: self.Close())
        self.dialoguepopup.Bind(wx.EVT_LEFT_DOWN, self.OnMouseButton)
        self.dialoguepopup.Bind(wx.EVT_LEFT_UP, self.OnMouseButton)
        self.dialoguepopup.Bind(wx.EVT_RIGHT_DOWN, self.OnMouseButton)
        self.dialoguepopup.Bind(wx.EVT_RIGHT_UP, self.OnMouseButton)
        self.dialoguepopup.Bind(wx.EVT_MIDDLE_DOWN, self.OnMouseButton)
        self.dialoguepopup.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.dialoguepopup.Bind(wx.EVT_KEY_UP, self.OnKeyUp)
        self.dialoguepopup.Bind(wx.EVT_MOTION, self.OnMouseMoved)
        self.dialoguepopup.Bind(wx.EVT_PAINT, self.OnPaint)

        self.system_flag = 0x0E0000
        self.music_enabled = True
        self.voice_enabled = True
        self.states = Interpreter(self, self)
        self.states.debug = self.debug
        self.debug.globalspanel.SetDataSource(self.states.globals)
        self.framebuffer = np.zeros((480, 640), dtype=np.uint8)
        self.frameregion = wx.Region()
        zf = zipfile.ZipFile(os.path.join(os.path.dirname(__file__), 'sognaresources.zip'), 'r')
        def resbmp(name):
            return wx.BitmapFromImage(wx.ImageFromStream(StringIO(zf.open(name).read())))
        self.carets = {
            1: [(resbmp('keybrd%d.bmp'%i), 6) for i in (1, 2)], # defaults from CTR
            2: [(resbmp('mouse%d.bmp'%i), 2) for i in xrange(1,9)]
        }
        self.default_font = wx.FontFromPixelSize((8, 16), wx.FONTFAMILY_SWISS, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, face=u'ＭＳ ゴシック')
        self.win_font = None

        self.audiolock = Lock()
        self.audioclips = [Sound() for i in xrange(4)]
        self.pcmformat = pyaudio.paInt8
        self.pcmchannels = 1
        self.loop_bgm = False
        self.playing_bgm = False
        self.bgmwav = None
        self.bgmmidi = False
        self.audiobuffer = np.zeros(17640, dtype=np.uint8)
        self.audiobuffer.dtype=np.int16
        self.audiowork1 = np.zeros(len(self.audiobuffer), dtype=np.int32)
        self.audiowork2 = np.zeros(len(self.audiobuffer), dtype=np.int32)
        self.audiostream = paud.open(format=pyaudio.paInt16, channels=2, rate=22050, output=True, stream_callback=self.audiocallback) # should be 22050...
        self.setvolumes()

        self.Bind(wx.EVT_TIMER, self.OnAnimationTimer)
        self.Bind(EVT_NEXT_STATE, self.OnNextState)
        self.Bind(EVT_SOUND_FINISHED, self.OnSoundFinished)
        self.t1 = wx.Timer(self)
        self.lasttime = time.time()
        self.times = deque([], 25)
        self.times2 = deque([], 25)

        self.SetClientSize((640, 480))
        self.CentreOnScreen()

    if sys.platform == 'win32':
        def MyWndProc(self, hWnd, msg, wParam, lParam):
            if msg == MM_MCINOTIFY:
                if wParam == 1 and self.bgmmidi:
                    if self.loop_bgm:
                        self.playmidibgm()
                    else:
                        self.stopbgm()
                return 0
            else:
                return CallWindowProc(self.oldWndProc, hWnd, msg, wParam, lParam)

    def loadwin(self, name):
        if not name.endswith('.WIN'):
            name += '.WIN'
        return bytecode.parse(self.dat.read(name))

    def loadanm(self, name):
        return Animation.parse(self.dat.read(name+'.ANM'))

    def loadbmp(self, name):
        return self.dat.read(name+'.BMP')

    def saveglobals(self, name, data):
        self.config.set('Saves', name, ' '.join(map(str, data)))
        self.SaveConfig()

    def loadglobals(self, name):
        try: s = self.config.get('Saves', name)
        except Exception: return
        if s: return map(int, s.split())

    def enableaction(self, action, enable=True):
        if action == 'repeat':
            self.repeatitem.Enable(enable)
        elif action == 'exit':
            self.restartitem.Enable(enable)
        elif action == 'save':
            self.menubar.EnableTop(self.menubar.FindMenu("Save"), enable)
        elif action == 'load':
            self.menubar.EnableTop(self.menubar.FindMenu("Load"), enable)
        else:
            incomplete('actions')
            print action, enable

    def setfullscreen(self, enable):
        incomplete('fullscreen')

    def OnUpdateVolumes(self, event):
        with MixerDialog(self, self, title='Volume') as dlg:
            dlg.CentreOnParent()
            dlg.ShowModal()

    def SaveSaveData(self, n):
        self.states.saveglobals('SaveData%d'%n, 16, 256)
        self.system_flag |= 1 << n
        self.loaditems[n].Enable(True)

    def LoadSaveData(self, n):
        self.states.loadglobals('SaveData%d'%n, 16, 256)
        self.states.setaddress(self.states.loadhandler)
        self.states.resetandnext()

    def SaveSystemVars(self):
        self.config.set('System', 'SystemFlag', self.system_flag)
        self.config.set('System', 'VoiceVolume', self.states.voice_volume)
        self.config.set('System', 'MusicVolume', self.states.music_volume)
        # TODO Language
        self.config.set('System', 'Name', self.states.hero_name.encode('utf_8'))
        self.SaveConfig()

    def LoadSystemVars(self):
        self.system_flag = self.config.getint('System', 'SystemFlag')
        for i, loaditem in enumerate(self.loaditems):
            loaditem.Enable(self.system_flag & 1 << i)
        self.musicitem.Check(self.system_flag & 0x020000)
        self.voiceitem.Check(self.system_flag & 0x040000)
        self.music_enabled = self.musicitem.IsChecked()
        self.voice_enabled = self.voiceitem.IsChecked()
        self.states.voice_volume = self.config.getint('System', 'VoiceVolume')
        self.states.music_volume = self.config.getint('System', 'MusicVolume')
        self.setvolumes()
        self.states.hero_name = self.config.get('System', 'Name').decode('utf_8')
        self.pcmformat = self.config.getint('System', 'PCMSampleWidth')
        self.pcmaudio8.Check(False)
        self.pcmaudio16.Check(False)
        if self.pcmformat == pyaudio.paInt16:
            self.pcmaudio16.Check(True)
        elif self.pcmformat == pyaudio.paInt8:
            self.pcmaudio8.Check(True)
        self.pcmchannels = self.config.getint('System', 'PCMChannels')
        self.pcmimmediate.Check(self.config.getint('System', 'PCMImmediate'))

    def ResetConfig(self):
        self.config = RawConfigParser()
        self.config.optionxform = str
        self.config.add_section('Saves')
        self.config.add_section('System')
        self.config.set('System', 'SystemFlag', self.system_flag)
        self.config.set('System', 'VoiceVolume', self.states.voice_volume)
        self.config.set('System', 'MusicVolume', self.states.music_volume)
        self.config.set('System', 'Name', self.states.hero_name.encode('utf_8'))
        self.config.set('System', 'PCMSampleWidth', self.pcmformat)
        self.config.set('System', 'PCMChannels', self.pcmchannels)
        self.config.set('System', 'PCMImmediate', int(self.pcmimmediate.IsChecked()))

    def ReloadConfig(self):
        self.ResetConfig()
        self.config.read(self.cfgpath)
        self.LoadSystemVars()

    def OnConfigControlUpdated(self, event):
        wxid = event.GetId()
        if wxid == self.pcmaudio8.GetId():
            self.pcmformat = pyaudio.paInt8
            self.config.set('System', 'PCMSampleWidth', pyaudio.paInt8)
        elif wxid == self.pcmaudio16.GetId():
            self.pcmformat = pyaudio.paInt16
            self.config.set('System', 'PCMSampleWidth', pyaudio.paInt16)
        elif wxid == self.pcmimmediate.GetId():
            self.config.set('System', 'PCMImmediate', int(self.pcmimmediate.IsChecked()))
        elif wxid == self.musicitem.GetId():
            if self.musicitem.IsChecked():
                self.system_flag |= 0x020000
                self.music_enabled = True
                if self.playing_bgm:
                    self.playbgm(self.current_bgm, self.loop_bgm)
            else:
                self.system_flag &= ~0x020000
                self.music_enabled = False
                if self.playing_bgm:
                    self.closebgm()
            self.SaveSystemVars()
        elif wxid == self.voiceitem.GetId():
            if self.voiceitem.IsChecked():
                self.system_flag |= 0x040000
                self.voice_enabled = True
            else:
                self.system_flag &= ~0x040000
                self.voice_enabled = False
                if self.states.current_sound is not None:
                    self.deletepcm(1)
            self.SaveSystemVars()

    def SaveConfig(self):
        self.config.write(open(self.cfgpath, 'w'))

    def OnExit(self, event):
        self.app.GetTopWindow().Close()

    def OnClose(self, event):
        with wx.MessageDialog(self, 'Return to Windows?', DEFAULT_CAPTION, wx.ICON_QUESTION|wx.YES_NO|wx.NO_DEFAULT) as dlg:
            if dlg.ShowModal() != wx.ID_YES:
                return
        try:
            self.SaveSystemVars()
        finally:
            self.audiostream.close()
            self.Destroy()

    def OnRestart(self, event):
        with wx.MessageDialog(self, 'Return to the menu?', DEFAULT_CAPTION, wx.ICON_QUESTION|wx.YES_NO|wx.NO_DEFAULT) as dlg:
            if dlg.ShowModal() != wx.ID_YES:
                return
        states = self.states
        states.setaddress(states.exithandler)
        states.resetandnext()

    def OnRepeat(self, event):
        states = self.states
        states.setaddress(states.repeathandler)
        states.resetandnext()

    def OnMove(self, event):
        x, y = self.ClientToScreenXY(0, 0)
        st = self.states
        self.dialoguepopup.SetDimensions(x+st.dlgpop_x, y+st.dlgpop_y, st.dlgpop_w, st.dlgpop_h)

    def OnMouseButton(self, event):
        states = self.states
        if event.ButtonDown():
            if event.LeftDown():
                states.inputflags |= INPUT_SKIP
                states.buttonflags |= INPUT_SKIP
            elif event.RightDown():
                states.inputflags |= INPUT_ESCAPE
                states.buttonflags |= INPUT_ESCAPE
                self.bmppanel.CaptureMouse()
            elif event.MiddleDown() and states.use_dlgpop == 2:
                self.toggledialoguepopup()
                return
            states.skip_type = 0
            self.InputDown()
        else:
            if event.LeftUp():
                states.inputflags &= ~INPUT_SKIP
                states.buttonflags &= ~INPUT_SKIP
            elif event.RightUp():
                states.inputflags &= ~INPUT_ESCAPE
                states.buttonflags &= ~INPUT_ESCAPE
                if self.bmppanel.HasCapture():
                    self.bmppanel.ReleaseMouse()
            self.InputUp()

    def OnKeyDown(self, event):
        code = event.GetKeyCode()
        states = self.states
        states = self.states
        if code in (wx.WXK_RETURN, wx.WXK_CONTROL, wx.WXK_F11, wx.WXK_F12):
            if code in (wx.WXK_CONTROL, wx.WXK_F11):
                states.skip_type = 1
            elif code == wx.WXK_F12:
                states.skip_type = 2
            states.inputflags |= INPUT_SKIP
        elif code == wx.WXK_SPACE:
            if states.use_dlgpop == 2:
                self.toggledialoguepopup()
            else:
                states.inputflags |= INPUT_SKIP
        elif code in (wx.WXK_ESCAPE, wx.WXK_NUMPAD0):
            states.inputflags |= INPUT_ESCAPE
        elif code == wx.WXK_F1:
            self.Iconize()
        elif code in (wx.WXK_UP, wx.WXK_NUMPAD8):
            states.inputflags |= INPUT_UP
        elif code in (wx.WXK_DOWN, wx.WXK_NUMPAD2):
            states.inputflags |= INPUT_DOWN
        elif code in (wx.WXK_LEFT, wx.WXK_NUMPAD4):
            states.inputflags |= INPUT_LEFT
        elif code in (wx.WXK_RIGHT, wx.WXK_NUMPAD6):
            states.inputflags |= INPUT_RIGHT
        elif code == wx.WXK_SHIFT:
            states.inputshift = 0x100

        self.InputDown()

        if states.choices_count:
            self.movechoice()
        if states.state == STATE_13:
            states.globals[states.typing_var] = states.inputshift + code
            self.clearstate()

        event.Skip()

    def OnKeyUp(self, event):
        code = event.GetKeyCode()
        states = self.states
        if code in (wx.WXK_RETURN, wx.WXK_SPACE, wx.WXK_CONTROL, wx.WXK_F11, wx.WXK_F12):
            if code == wx.WXK_CONTROL:
                states.skip_type = 0
            states.inputflags &= ~INPUT_SKIP
        elif code in (wx.WXK_ESCAPE, wx.WXK_NUMPAD0):
            states.inputflags &= ~INPUT_ESCAPE
        elif code in (wx.WXK_UP, wx.WXK_NUMPAD8):
            states.inputflags &= ~INPUT_UP
        elif code in (wx.WXK_DOWN, wx.WXK_NUMPAD2):
            states.inputflags &= ~INPUT_DOWN
        elif code in (wx.WXK_LEFT, wx.WXK_NUMPAD4):
            states.inputflags &= ~INPUT_LEFT
        elif code in (wx.WXK_RIGHT, wx.WXK_NUMPAD6):
            states.inputflags &= ~INPUT_RIGHT
        elif code == wx.WXK_SHIFT:
            states.inputshift = 0
        else:
            event.Skip()
            return
        self.InputUp()

    def InputDown(self):
        states = self.states
        inputflags, state = states.inputflags, states.state
        if not inputflags & (INPUT_ESCAPE | INPUT_SKIP):
            if inputflags and state == STATE_10:
                states.globals[states.input_var] = inputflags
                self.clearstate()
        elif not inputflags & INPUT_ESCAPE or not states.mouse_caret:
            if states.choices_count:
                if inputflags & INPUT_ESCAPE:
                    states.globals[states.choices_var] = -1
                self.clearchoices(True)
            if states.continuehandler:
                states.setaddress(states.continuehandler)
                states.resetandnext()
                return
            if inputflags & INPUT_ESCAPE and states.escapehandler:
                states.setaddress(states.escapehandler)
                states.resetandnext()
                return
            if state == STATE_1:
                self.clearcaret()
                self.clearstate()
                return
            if state == STATE_7 and not states.noskipsound:
                self.deletepcm(1)
                return
            if state == STATE_8:
                states.reading_countdown = 0
                self.clearstate()
            if inputflags and state == STATE_10:
                states.globals[states.input_var] = inputflags
                self.clearstate()

    def InputUp(self):
        states = self.states
        if not states.inputflags and states.state == STATE_11:
            self.clearstate()

    def OnMouseMoved(self, event):
        obj = event.GetEventObject()
        states = self.states
        mx, my = event.GetPositionTuple()
        states.mouse_pos_new = (mx, my)

        if self.dialoguepopup != obj:
            if states.mouse_caret and event.RightIsDown(): # limited to right button, original took anything
                if states.mouse_scrolling:
                    mxold, myold = states.mouse_pos_old
                    x, x2, y, y2 = states.scroll_x1, states.scroll_x2, states.scroll_y1, states.scroll_y2
                    if x2: states.scroll_x1 = x = min(max(0, x + mxold - mx), x2)
                    if y2: states.scroll_y1 = y = min(max(0, y + myold - my), y2)
                    for al in states.anim_layers:
                        if al.flags & FLAG_MOTION:
                            if x2: al.vector_curx = x * al.vector_deltax
                            if y2: al.vector_cury = y * al.vector_deltay
                else:
                    states.mouse_scrolling = True
            else:
                states.mouse_scrolling = False

            if states.mouse_pos_new != states.mouse_pos_old:
                states.mouse_moved = True
            states.mouse_pos_old = states.mouse_pos_new

        if states.choices_count and self.getdialoguetarget() == obj:
            choice = None
            crect = wx.Rect(states.choices_x, states.choices_y, states.choices_w, states.choices_h)
            cellw, fonth, layout = states.choices_cellw, states.font_height, states.choices_layout
            if crect.ContainsXY(mx, my):
                if layout < 1:
                    choice = (my - crect.Top) / fonth
                elif layout == 1:
                    choice = (mx - crect.Left) / cellw
                elif layout == 3:
                    choice = 3 * (my - crect.Top) / fonth + (mx - crect.Left) / cellw
                elif layout == 4:
                    choice = 4 * (my - crect.Top) / fonth + (mx - crect.Left) / cellw
            if choice is not None and choice != states.globals[states.choices_var]:
                states.globals[states.choices_var] = choice
                self.paintchoices(False)

    def clearstate(self):
        self.states.state = 0
        wx.PostEvent(self, NextStateEvent())

    def cleararea(self, x, y, w, h):
        self.framebuffer[y:y+h, x:x+w] = 0
        self.frameregion.Union(x, y, w, h)
        self.repaint()
        self.frameregion.Clear()

    def setfontheight(self, height):
        self.win_font = wx.FontFromPixelSize((height / 2, height), wx.FONTFAMILY_SWISS, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, face=u'ＭＳ ゴシック')

    def SetDatPath(self, path):
        self.datpath = path
        self.dat = dat = DATArchive(path)
        self.basepath = os.path.dirname(path)
        self.cfgpath = cfgpath = path+'.pysogna.cfg'
        self.ReloadConfig()
        self.debug.SetWinNames([winname for winname in dat.filenames() if winname.endswith('.WIN')])
        self.SetWinName('START.WIN')

    def SetWinName(self, name):
        self.LoadWin(name)

    def SetStartAddress(self, address):
        self.states.index = self.states.addresses[address]
        self.clearstate()

    def LoadWin(self, name):
        win = bytecode.parse(self.dat.read(name))
        self.states.setops(win)
        self.debug.setcurrentwin(name[:-4], win)
        self.clearstate()

    def OnNextState(self, event):
        self.states.nextstate()

    def OnSoundFinished(self, event):
        if event.n == 0:
            self.storebgmchunk()
        elif event.n == 1:
            states = self.states
            states.current_sound = None
            if states.state == STATE_7:
                self.clearstate()
        
    def StartGame(self):
        self.t1.Start(32)

    def StopGame(self):
        self.t1.Stop()

    def IsGameRunning(self):
        return self.t1.IsRunning()

    def repaint(self):
        region = self.frameregion
        if not region.IsEmpty():
            frame = self.framebuffer
            palette = self.states.bitmap_palette
            data = palette[frame.reshape(-1,)].reshape(frame.shape+(3,))
            bufferbmp = wx.BitmapFromBuffer(640, 480, data)
            dc = wx.ClientDC(self.bmppanel)
            dc.SetClippingRegionAsRegion(region)
            dc.DrawBitmap(bufferbmp, 0, 0)
            dc.DestroyClippingRegion()

    def OnPaint(self, event):
        obj = event.GetEventObject()
        dc = wx.PaintDC(obj)
        if obj == self.bmppanel:
            frame = self.framebuffer
            palette = self.states.bitmap_palette
            data = palette[frame.reshape(-1,)].reshape(frame.shape+(3,))
            bufferbmp = wx.BitmapFromBuffer(640, 480, data)
            dc.DrawBitmap(bufferbmp, 0, 0)
        else:
            self.paintdialoguepopup(dc=dc)
            self.bmppanel.SetFocus()
        if obj == self.getdialoguetarget(lt=False):
            self.redrawdialogue(dc)
            if self.states.choices_count:
                self.paintchoices(dc=dc)

    def OnAnimationTimer(self, event):
        states = self.states
        if states.pause_anim or self.IsIconized():
            return

        starttime = time.time()

        drawrect = wx.Rect(states.drawing_x, states.drawing_y, states.drawing_w, states.drawing_h)
        region = self.frameregion

        if states.palette_changed:
            states.changepalette()
            region.UnionRect(drawrect)

        self.repaint()

        if states.state == STATE_5:
            self.clearstate()
        if states.delay_countdown:
            states.delay_countdown -= 1
            if not states.delay_countdown and states.state == STATE_2:
                self.clearstate()
        if states.reading_countdown:
            states.reading_countdown -= 1
            if not states.reading_countdown and states.state == STATE_8:
                self.clearstate()

        region.Clear()

        if states.caret_type:
            self.paintcaret()

        if states.state == STATE_15:
            self.drawmap(states.mapvar);

        anim_layers = states.anim_layers
        for al_i in xrange(32):
            if not anim_layers[al_i].flags & 0x8000:
                if states.state == STATE_3 and states.waitforlayer == al_i:
                    self.clearstate()
                    states.pause_anim = states.future_pause
                    states.future_pause = False
                if states.typing_layer == al_i and states.state == STATE_13:
                    states.globals[states.typing_var] = 0
                    self.clearstate()
                    break
                continue

            al = anim_layers[al_i]
            al.frame_count -= 1
            while not al.frame_count:
                if al.flags & FLAG_RANDOM:
                    for i in xrange(randint(0, 3)):
                        al.sequence_index += 1
                        if al.sequence_index >= len(al.sequences):
                            al.sequence_index = 0

                i = al.sequence_index
                al.sequence_index = i + 1
                seq = al.sequences[i]
                al.frame_count = seq.count
                al.image_index = image_index = seq.index - 1
                if 0 <= image_index < 128:
                    loaded_frames = states.loaded_frames[al.anm_index]
                    image = withmask(loaded_frames.offsets[image_index].frame.value)

                    if al.flags & FLAG_WAVE:
                        viewrect = wx.Rect(al.sequence_x + image.x, al.sequence_y + image.y, al.vector_width, al.vector_height)
                        region.UnionRect(viewrect)
                        x, y, w, h = viewrect
                        imgto = self.framebuffer[y:y+h, x:x+w]
                        x, y = al.vector_curx >> 16, al.vector_cury >> 16
                        imgfr = image.nparray[y:y+h]
                        xx = np.arange(viewrect.Height); xx *= al.wave_deltax; xx += al.wave_shift; xx &= 0xFF
                        xx = COSINES[xx]; xx *= al.wave_cur; xx /= 10000; xx += x
                        # TODO look into making this safer
                        for y, x in enumerate(xx):
                            imgto[y] = imgfr[y, x:x+w]

                        al.wave_shift = (al.wave_shift + al.wave_deltay) & 0xFF # was 0x7ff, see comment in interpreter
                        if al.wave_countdown:
                            al.wave_countdown -= 1
                            al.wave_cur += al.wave_delta
                        elif states.state == STATE_12 and states.waitforlayer == al_i:
                            self.clearstate()

                    else:
                        x, y = al.sequence_x + image.x, al.sequence_y + image.y
                        if al.flags & FLAG_MOTION:
                            w, h = al.vector_width, al.vector_height # and this is why tweens are still specified for points
                            if al.flags & FLAG_MOTION_POINT:
                                if al.points:
                                    dx, dy = al.points[al.current_point]
                                else:
                                    dx, dy = 0, 0
                            else:
                                dx, dy = al.vector_curx >> 16, al.vector_cury >> 16
                        else:
                            w, h, dx, dy = image.w, image.h, 0, 0

                        imgrect = wx.Rect(0, 0, image.w, image.h) # image to be shown
                        viewrect = wx.Rect(dx, dy, w, h) # region that we want
                        viewrect = viewrect.Intersect(imgrect) # region that we want that exists
                        x, y = x - dx, y - dy
                        dx, dy = viewrect.Left, viewrect.Top # replace in case the top or left is out of bounds
                        viewrect.OffsetXY(x, y) # moved to its intended location
                        viewrect = viewrect.Intersect(drawrect) # and restrict to the drawing area just in case
                        x, y, w, h = viewrect

                        imgto = self.framebuffer[y:y+h, x:x+w]
                        x, y = dx, dy
                        imgfr = image.nparray[y:y+h, x:x+w]

                        # setup indexes (shimmer effect and optional)
                        index = Ellipsis
                        if al.flags & FLAG_SHIMMER:
                            # not sure this is identical, but it looks close enough
                            m = states.shimmer_value
                            n = m << 20 & 0xF0F00000 | m << 16 & 0x0F0F0000 | m >> 4 & 0x00000F0F | m & 0x0000F0F0
                            w2 = (w+3)/4*4
                            index = np.zeros((h, w2), dtype=bool)
                            for i in xrange(h-1,-1,-1):
                                for j in xrange(0, w2, 4):
                                    index[i, j:j+4] = [n & 0x8000, n & 0x4000, n & 0x2000, n & 0x8000]
                                    n = n & ~0xFFFF | n << 8 & 0xFF00 | n >> 7 & 0xFE | n >> 15 & 0x01
                                n = n >> 16 & 0x0000FFFF | n << 16 & 0xFFFF0000
                            if w != w2:
                                index = index[:, :w]

                            if states.shimmer_counter:
                                if states.shimmer_type == 1:
                                    states.shimmer_value = m & ~0xFFFF | m >> 1 & 0x7F7F | m << 7 & 0x8080 | 0x8008
                                elif states.shimmer_type == 2:
                                    m = m & ~0xFF00 | m >> 1 & 0x7F00
                                    m = m & ~0xFF | m >> 12 & 0x0F | m >> 4 & 0xF0
                                    states.shimmer_value = m
                                states.shimmer_counter -= 1

                            if al.flags & FLAG_TRANSPARENCY:
                                index &= image.masks[al.mask_color][y:y+h, x:x+w]

                        elif al.flags & FLAG_TRANSPARENCY:
                            index = image.masks[al.mask_color][y:y+h, x:x+w]

                        # copy
                        try:
                            imgto[index] = imgfr[index]

                        except Exception:
                            print_exc()
                            print >>sys.stderr, viewrect, dx, dy, self.framebuffer.shape, image.nparray.shape
                            print >>sys.stderr, al
                            al.flags = 0

                        else:
                            region.UnionRect(viewrect)

                if al.flags & FLAG_MOTION:
                    if al.flags & FLAG_MOTION_POINT:
                        if al.point_countdown:
                            al.point_countdown -= 1
                            if not al.point_countdown:
                                if al.flags & FLAG_MOTION_LOOP:
                                    al.point_countdown = al.point_count
                                    al.current_point = 0
                            else:
                                al.current_point += 1
                        elif states.state == STATE_6 and states.waitforlayer == al_i:
                            self.clearstate()
                    elif al.vector_countdown:
                        al.vector_curx += al.vector_deltax
                        al.vector_cury += al.vector_deltay
                        al.vector_countdown -= 1
                        if not al.vector_countdown and al.flags & FLAG_MOTION_LOOP:
                            if al.flags & FLAG_MOTION_BOUNCE:
                                al.vector_deltax = -al.vector_deltax
                                al.vector_deltay = -al.vector_deltay
                            else:
                                al.vector_curx = al.vector_initx
                                al.vector_cury = al.vector_inity
                            al.vector_countdown = al.vector_initcount
                    elif states.state == STATE_6 and states.waitforlayer == al_i:
                        self.clearstate()

            if al.frame_count == 1 and al.sequence_index >= len(al.sequences):
                if al.flags & FLAG_LOOP:
                    al.sequence_index = 0
                else:
                    if states.state == STATE_3 and states.waitforlayer == al_i:
                        self.clearstate()
                        states.pause_anim = states.future_pause
                        states.future_pause = 0
                    if al.flags & FLAG_REPEATLAST:
                        al.sequence_index -= 1
                    else:
                        anim_layers[al_i].flags = 0

        if not region.IsEmpty():
            if states.future_palette_changed:
                states.palette_changed = True
                states.future_palette_changed = False

        if states.music_delta:
            states.music_volume2 += states.music_delta
            if states.music_delta <= 0:
                if states.music_volume2 < states.music_stepto:
                    states.music_volume2 = states.music_stepto
                    states.music_delta = 0
            else:
                if states.music_volume2 > states.music_stepto:
                    states.music_volume2 = states.music_stepto
                    states.music_delta = 0
            self.setvolumes()

        t = time.time()
        times, times2 = self.times, self.times2
        times.append(t - self.lasttime)
        times2.append(t - starttime)
        self.lasttime = t
        self.debug.fps.SetValue('%5.2f %5.4f'%(1/(sum(times)/len(times)), sum(times2)/len(times2)))

    # dialogue

    def drawdialogue(self, draw=True):
        states = self.states
        target = self.getdialoguetarget()
        if target is None:
            return
        x, y, w, h = states.text_char_x, states.text_char_y, states.text_char_w, states.text_char_h
        x_old, y_old = x, y
        line, text = states.dialogue_line, states.text_lines
        nlpadding, bytes_printed = states.text_nlpadding, states.bytes_printed

        i, j = 0, len(line)
        while i < j:
            c = line[i]
            i += 1
            if c == CHAR_TEXTSUBST:
                states.dialogue_line = states.names[ord(line[i])-1]
                i += 1
                states.text_char_x, states.text_char_y, states.bytes_printed = x, y, bytes_printed
                self.drawdialogue(False)
                x, y, bytes_printed = states.text_char_x, states.text_char_y, states.bytes_printed
            elif c == CHAR_TEXTCLEAR:
                text[:] = [[]]
                fh = states.font_height
                dc = wx.ClientDC(target)
                dc.SetPen(wx.BLACK_PEN)
                dc.SetBrush(wx.BLACK_BRUSH)
                dc.DrawRectangle(states.text_x, states.text_y, fh * w / 2, fh * h)
                states.text_char_x = 0;
                x_old = 0;
                x = 0;
                states.text_char_y = 0;
                y_old = 0;
                y = 0;
                states.linecount = 0;
            elif c == CHAR_TEXTCRLF:
                y += 1
                text.append([])
                x = 0
            elif c == CHAR_TEXTHERO:
                states.dialogue_line = states.hero_name
                states.text_char_x, states.text_char_y, states.bytes_printed = x, y, bytes_printed
                self.drawdialogue(False)
                x, y, bytes_printed = states.text_char_x, states.text_char_y, states.bytes_printed
            else:
                if not x and draw:
                    text[y].append(' '*nlpadding)
                    x += nlpadding
                cw = 1 if ord(c) <= 0x80 else 2
                if x > w - cw:
                    y += 1
                    text.append([])
                    text[y].append(' '*nlpadding)
                    x = nlpadding
                text[y].append(c)
                bytes_printed += cw
                x += cw

        if draw:
            tx, ty, fh = states.text_x, states.text_y, states.font_height
            dc = wx.ClientDC(target)
            if self.win_font:
                dc.SetFont(self.win_font)
            else:
                dc.SetFont(self.default_font)
            dc.SetTextForeground(wx.WHITE)
            dc.SetTextBackground(wx.BLACK)
            while y + 1 > y_old:                
                s = ''.join(text[y_old]) # ignoring x_old, going to draw the entire line over the prior draw.
                dc.DrawText(s, tx, ty + fh * y_old) # x is supposed to be: tx + (fh * x_old) / 2
                y_old += 1
                x_old = 0
            dc.SetFont(wx.NORMAL_FONT)

            states.reading_countdown = states.reading_factor * 2 * bytes_printed / 100

            if line:
                self.debug.dialogtext.SetValue(line)

        states.text_char_x = x
        states.text_char_y = y
        states.dialogue_line = ''

    def redrawdialogue(self, dc):
        states = self.states
        tx, ty, fh, text = states.text_x, states.text_y, states.font_height, states.text_lines
        if self.win_font:
            dc.SetFont(self.win_font)
        else:
            dc.SetFont(self.default_font)
        dc.SetTextForeground(wx.WHITE)
        dc.SetTextBackground(wx.BLACK)
        for i in xrange(states.text_char_y + 1):
            s = ''.join(text[i])
            dc.DrawText(s, tx, ty + fh * i)
        dc.SetFont(wx.NORMAL_FONT)

    def drawtext(self):
        states = self.states
        s, fh, tc = states.dialogue_line, states.font_height, states.text_color
        tx, ty = states.text_position
        for index, (r, g, b) in enumerate(states.current_palette):
            if (r, g, b) == tc:
                break
        else:
            # TODO find the nearest palette color
            # falling back to spray and pray
            dc = wx.ClientDC(self.bmppanel)
            if self.win_font:
                dc.SetFont(self.win_font)
            else:
                dc.SetFont(self.default_font)
            dc.SetTextForeground(states.text_color)
            dc.SetBackgroundMode(wx.TRANSPARENT)
            dc.DrawText(s, tx, ty)
            dc.SetFont(wx.NORMAL_FONT)

        w, h = fh*len(s.encode('sjis'))/2, fh
        tempbmp = wx.EmptyBitmap(w, h, 1)
        dc = wx.MemoryDC(tempbmp)
        if self.win_font:
            dc.SetFont(self.win_font)
        else:
            dc.SetFont(self.default_font)
        dc.SetTextForeground(wx.WHITE)
        dc.SetBackgroundMode(wx.TRANSPARENT)
        dc.DrawText(s, 0, 0)
        dc.SetFont(wx.NORMAL_FONT)
        dc.SelectObject(wx.NullBitmap)
        arr = np.zeros((h, w), dtype=np.uint32)
        tempbmp.CopyToBuffer(arr, wx.BitmapBufferFormat_RGB32)
        mask = arr == 0xFFFFFFFF
        self.framebuffer[ty:ty+h, tx:tx+w][mask] = index
        self.frameregion.Union(tx, ty, w, h)
        self.repaint()
        self.frameregion.Clear()

    # carets

    def paintcaret(self):
        states = self.states
        if states.caret_countdown:
            states.caret_countdown -= 1
        else:
            target = self.getdialoguetarget()
            carets = self.carets[states.caret_type]
            index = states.caret_index
            bmp, frames = carets[index]
            if target is not None:
                x, y = states.caret_pos
                dc = wx.ClientDC(target)
                dc.DrawBitmap(bmp, x, y)
            states.caret_countdown = frames
            index += 1
            if index >= len(carets):
                index = 0
            states.caret_index = index

    def clearcaret(self):
        states = self.states
        if states.caret_type:
            states.caret_type = 0
            x, y = states.caret_pos
            target = self.getdialoguetarget()
            if target is not None:
                dc = wx.ClientDC(target)
                dc.SetPen(wx.BLACK_PEN)
                dc.SetBrush(wx.BLACK_BRUSH)
                dc.DrawRectangle(x, y, 18, 18)

    # choices

    def gettextwidth(strings):
        raise NotImplementedError('choices')

    def paintchoices(self, background=False, dc=None):
        states = self.states
        x, y, w, h = states.choices_x, states.choices_y, states.choices_w, states.choices_h
        cellw, fonth, layout = states.choices_cellw, states.font_height, states.choices_layout
        if dc is None:
            target = self.getdialoguetarget()
            if target is None:
                return
            dc = wx.ClientDC(target)

        if background:
            dc.SetPen(wx.BLACK_PEN)
            dc.SetBrush(wx.BLACK_BRUSH)
            dc.DrawRectangle(x, y, w, h)

        if self.win_font:
            dc.SetFont(self.win_font)
        else:
            dc.SetFont(self.default_font)

        for i, text in enumerate(states.choices_text):
            dc.SetBackgroundMode(wx.SOLID)
            if i == states.globals[states.choices_var]:
                dc.SetTextForeground(wx.BLACK)
                dc.SetTextBackground(wx.WHITE)
            else:
                dc.SetTextForeground(wx.WHITE)
                dc.SetTextBackground(wx.BLACK)

            # if states.language >= 1: # same implementation
            if layout < 1:
                dc.DrawText(text, x, y + fonth * i)
            elif layout == 1:
                dc.DrawText(text, x + cellw * i, y)
            elif layout == 3:
                dc.DrawText(text, x + cellw * i % 3, y + fonth * i / 3)
            elif layout == 4:
                dc.DrawText(text, x + cellw * i % 4, y + fonth * i / 4)

        dc.SetFont(wx.NORMAL_FONT)

    def clearchoices(self, reset):
        states = self.states
        if states.choices_count:
            target = self.getdialoguetarget()
            if target is not None:
                x, y, w, h = states.choices_x, states.choices_y, states.choices_w, states.choices_h
                dc = wx.ClientDC(target)
                dc.SetPen(wx.BLACK_PEN)
                dc.SetBrush(wx.BLACK_BRUSH)
                dc.DrawRectangle(x, y, w, h)

            if reset:
                states.case105_2 = False
                states.choices_count = 0
                self.clearstate()

    def movechoice(self):
        st = self.states
        gl = st.globals
        i = st.choices_var
        flags = st.inputflags & 0xF
        if st.choices_layout == 0:
            if flags == INPUT_DOWN:
                if gl[i] < st.choices_count - 1: gl[i] += 1
                else: gl[i] = 0
            elif flags == INPUT_UP:
                if gl[i] > 0: gl[i] -= 1
                else: gl[i] = st.choices_count - 1
        elif st.choices_layout == 1:
            if flags == INPUT_RIGHT:
                if gl[i] < st.choices_count - 1: gl[i] += 1
                else: gl[i] = 0
            elif flags == INPUT_LEFT:
                if gl[i] > 0: gl[i] -= 1
                else: gl[i] = st.choices_count - 1
        elif st.choices_layout == 3:
            if flags == INPUT_UP:
                if gl[i] >= 3: gl[i] -= 3
            elif flags == INPUT_DOWN:
                if gl[i] < st.choices_count - 3: gl[i] += 3
            elif flags == INPUT_LEFT:
                if gl[i] > 0: gl[i] -= 1
            elif flags == INPUT_RIGHT:
                if gl[i] < st.choices_count - 1: gl[i] += 1
        elif st.choices_layout == 4:
            if flags == INPUT_UP:
                if gl[i] >= 4: gl[i] -= 4
            elif flags == INPUT_DOWN:
                if gl[i] < st.choices_count - 4: gl[i] += 4
            elif flags == INPUT_LEFT:
                if gl[i] > 0: gl[i] -= 1
            elif flags == INPUT_RIGHT:
                if gl[i] < st.choices_count - 1: gl[i] += 1
        self.paintchoices(False)

    # audio

    def setvolumes(self):
        states = self.states
        music_volume = states.music_volume2 * states.music_volume / 0xFFFF
        self.setpcmvolume(0, music_volume >> 8)
        self.setpcmvolume(1, states.voice_volume >> 8)
        self.setmidivolume(music_volume)

    def setpcmvolume(self, n, vol):
        self.audioclips[n].vol1 = vol

    def setmidivolume(self, vol):
        # # The below changes the application volume, not just MIDI
        # winmm.midiOutSetVolume(0, vol & 0xFFFF | vol << 16 & 0xFFFF0000)
        pass

    def loadpcm(self, name, loop):
        if self.voice_enabled:
            data = self.dat.read(name+'.PCM')
            data = np.frombuffer(data, dtype=np.uint8)
            self.states.current_sound = data
            self.storepcm(1, data, loop)

    def storepcm(self, n, data, loop):
        audioclips, lock = self.audioclips, self.audiolock
        lock.acquire()
        try:
            clip = audioclips[n]
            if self.pcmimmediate.IsChecked():
                # workaround for older SGS - though SGS seems to actually stop the sound earlier
                clip.data = None
                clip.nextdata = None
            if clip.data is None:
                clip.data = data
                clip.index = 0
                clip.loop = loop
                clip.channels = self.pcmchannels
                clip.format = self.pcmformat
            elif clip.nextdata is None:
                clip.nextdata = (data, self.pcmchannels, self.pcmformat)
        finally:
            lock.release()

    def composepcm(self, name, pos, merge):
        if self.voice_enabled:
            data = self.dat.read(name+'.PCM')
            data = np.frombuffer(data, dtype=np.uint8)
            if merge:
                sound = self.states.current_sound
                if pos + len(data) > len(sound):
                    temp = np.zeros(pos + len(data), dtype=np.uint8)
                    temp[:len(sound)] = sound
                    sound = temp
                if self.pcmformat == pyaudio.paInt16:
                    dstview = sound.view('int16')
                    srcview = data.view('int16')
                    minv, maxv = -32768, 32767
                    pos /= 2
                elif self.pcmformat == pyaudio.paInt8:
                    dstview = sound.view('int8')
                    srcview = data.view('int8')
                    minv, maxv = -128, 127
                else: raise NotImplementedError('pcm format', self.pcmformat)
                work = dstview.astype(dtype=np.int32)
                work[pos:pos+len(srcview)] += srcview
                np.clip(work, minv, maxv, dstview)
                self.states.current_sound = sound
            elif pos:
                sound = np.zeros(pos + len(data), dtype=np.uint8)
                sound[pos:] = data
                self.states.current_sound = sound
            else:
                self.states.current_sound = data

    def startpcm(self, loop):
        if self.voice_enabled:
            sound = self.states.current_sound 
            if sound is not None:
                self.storepcm(1, sound, loop)

    def deletepcm(self, n):
        audioclips, lock = self.audioclips, self.audiolock
        lock.acquire()
        try:
            if audioclips[n].data is not None:
                audioclips[n].data = None
                audioclips[n].nextdata = None
                wx.PostEvent(self, SoundFinishedEvent(n=n))
        finally:
            lock.release()

    def playbgm(self, n, loop):
        self.current_bgm = n
        self.loop_bgm = loop
        self.playing_bgm = True
        if self.music_enabled:
            name = self.states.track_names[n]
            if name:
                path = os.path.join(self.basepath, name)
                path2 = path + '.WAV'
                if os.path.exists(path2):
                    self.loadwavbgm(path2)
                    return
                path2 = path + '.MID'
                if os.path.exists(path2):
                    self.loadmidibgm(path2)

    def loadwavbgm(self, path):
        self.bgmwav = wave.open(path, 'rb')
        self.storebgmchunk()
        self.storebgmchunk()

    def storebgmchunk(self):
        clip, wav, lock = self.audioclips[0], self.bgmwav, self.audiolock
        if wav is None:
            return
        frames = 44100 # TODO put this somewhere?
        data = wav.readframes(frames)
        if not data:
            if self.loop_bgm:
                wav.rewind()
                data = wav.readframes(frames)
            else:
                self.closebgm()
                return
        data = np.frombuffer(data, dtype=np.uint8)

        lock.acquire()
        try:
            channels = wav.getnchannels()
            format = paud.get_format_from_width(wav.getsampwidth())
            if clip.data is not None:
                clip.nextdata = (data, channels, format)
            else:
                clip.data = data
                clip.index = 0
                clip.loop = False
                clip.channels = channels
                clip.format = format
        finally:
            lock.release()

    def loadmidibgm(self, path):
        if self.bgmmidi:
            self.closebgm()
        res = mciSendString('open "{0}" type sequencer alias bgm'.format(path))
        if mciSendString('status bgm port') == 'mapper':
            self.bgmmidi = True
            self.playmidibgm()
        else:
            mciSendString('close bgm')

    def playmidibgm(self):
        mciSendString('play bgm from 0 notify', hwnd=self.GetHandle())

    def stopbgm(self):
        self.closebgm()
        self.playing_bgm = False

    def closebgm(self):
        if self.bgmmidi:
            mciSendString('stop bgm')
            mciSendString('close bgm')
            self.bgmmidi = False
        self.bgmwav = None
        self.deletepcm(0)

    def audiocallback(self, in_data, frame_count, time_info, status_flags):
        lock, clips, buffer, work1, work2 = self.audiolock, self.audioclips, self.audiobuffer, self.audiowork1, self.audiowork2
        lock.acquire()
        try:
            work2[...] = 0
            for n, clip in enumerate(clips):
                work1[...] = 0
                index, data = clip.getdata()
                count = 0
                while count < frame_count and data is not None:
                    length = 0 if data is None else (len(data) - index) / clip.channels # length in frames
                    if count + length < frame_count:
                        a, b = 2*count, 2*(count+length)
                        if clip.channels == 1:
                            work1[a:b:2] = data[index:]
                            work1[a+1:b:2] = data[index:]
                        elif clip.channels == 2:
                            work1[a:b] = data[index:]
                        count += length
                        index, data = clip.loopornext()
                        wx.PostEvent(self, SoundFinishedEvent(n=n))
                    else:
                        a, b = 2*count, 2*frame_count
                        if clip.channels == 1:
                            c, d = index, index+frame_count-count
                            work1[a:b:2] = data[c:d]
                            work1[a+1:b+1:2] = data[c:d]
                        elif clip.channels == 2:
                            c, d = index, index+2*(frame_count-count)
                            work1[a:b] = data[c:d]
                        count = frame_count
                        index = d
                    if clip.format == pyaudio.paInt8:
                        work1[a:b] <<= 8

                clip.index = index
                work1 *= clip.vol1 * clip.vol2
                work1 >>= 16
                work2 += work1
            np.clip(work2, -32768, 32767, out=buffer)
            return (self.audiobuffer[:2*frame_count], pyaudio.paContinue)
        finally:
            lock.release()

    def drawmap(self, map_var):
        st = self.states
        mgl = st.globals[map_var:]
        st.mapmove_orient = mgl[15]
        if mgl[0] != 1:
            if mgl[0] < 3:
                if mgl[0] == 0:
                    mgl[0] = 1
                img = st.loaded_frames[mgl[1]]
                img = img.offsets[0].frame.value
                st.map_width = img.w
                st.map_height = img.h
                st.walk_index = 0
                st.mapmove_x = 0
                st.mapmove_y = 0
                st.walk_sprite_index = 0
                x, y = mgl[4:6]
                for i in xrange(64):
                    move_trail = st.move_trail[63 - i]
                    if st.mapmove_orient == 0:
                        move_trail[0:2] = x, y - i
                    elif st.mapmove_orient == 1:
                        move_trail[0:2] = x, y + i
                    elif st.mapmove_orient == 2:
                        move_trail[0:2] = x + i, y
                    elif st.mapmove_orient == 3:
                        move_trail[0:2] = x - i, y
                    move_trail[2] = 3 * st.mapmove_orient
                st.trail_index = 0
                moved = True
            elif mgl[0] == 4:
                if mgl[17]:
                    mgl[17] -= 1
                    st.mapmove_orient = mgl[16]
                    if st.mapmove_orient == 0:
                        st.mapmove_orient, st.mapmove_x, st.mapmove_y = 1, 0, -7
                    elif st.mapmove_orient == 1:
                        st.mapmove_orient, st.mapmove_x, st.mapmove_y = 2, -7, 0
                    elif st.mapmove_orient == 2:
                        st.mapmove_orient, st.mapmove_x, st.mapmove_y = 0, 0, 7
                    elif st.mapmove_orient == 3:
                        st.mapmove_orient, st.mapmove_x, st.mapmove_y = 3, 7, 0
                    st.walk_index += 1
                    if st.walk_index == 12:
                        st.walk_index = 0
                    st.walk_sprite_index = st.walk_index / 3
                    moved = True
                else:
                    self.clearstate()

        elif st.inputflags & 0x2F:
            if st.buttonflags & 0x20:
                mx, my = st.mouse_pos_new
                y = float(mx - st.drawing_w / 2) # x and y are reversed...
                x = float(my - st.drawing_h / 2)
                if y + x == 0.0:
                    st.mapmove_x, st.mapmove_y = 0, 0
                    theta = 0
                else:
                    theta = math.atan2(y, x)
                    st.mapmove_x = int(math.sin(theta) * 8.0)
                    st.mapmove_y = int(math.cos(theta) * 8.0)
                theta = 2 * theta / math.pi + 2.5
                n = int(theta) % 4
                if n == 0:
                    st.mapmove_orient = 1
                elif n == 1:
                    st.mapmove_orient = 2
                elif n == 2:
                    st.mapmove_orient = 0
                elif n == 3:
                    st.mapmove_orient = 3
            else:
                temp = MAP_KEY_XYO.get(st.inputflags & 0xF, None)
                if temp is not None:
                    st.mapmove_x, st.mapmove_y, st.mapmove_orient = temp
            st.walk_index += 1
            if st.walk_index == 12:
                st.walk_index = 0
            st.walk_sprite_index = st.walk_index / 3
            moved = True
        elif st.walk_index:
            st.walk_index += 1
            if st.walk_index == 12:
                st.walk_index = 0
            st.walk_sprite_index = st.walk_index / 3
            if not st.walk_sprite_index or st.walk_sprite_index == 2:
                st.walk_index = 0
            moved = True
        else:
            st.mapmove_x, st.mapmove_y = 0, 0
            if mgl[12]:
                moved = True
            else:
                moved = False

        if moved:
            img = st.loaded_frames[mgl[2]].offsets[0].frame.value
            def colorat(dx, dy):
                x, y = mgl[4] + dx, mgl[5] + dy
                i = img.w * (img.h - y / 16 - 1) + x / 16
                # i = st.map_width * (st.map_height - y - 16) / 256 + x / 16
                return ord(img.data[i])
            color = colorat(st.mapmove_x, st.mapmove_y)
            if color != 12:
                if st.mapmove_x and colorat(st.mapmove_x, 0) == 12:
                    st.mapmove_y = 0
                elif st.mapmove_y and colorat(0, st.mapmove_y) == 12:
                    st.mapmove_x = 0
                else:
                    if color <= 10:
                        for i in xrange(1, 1 if st.mapmove_x else 8):
                            if colorat(i, st.mapmove_y) == 12:
                                st.mapmove_x, st.mapmove_y = 7, 0
                                break
                            if colorat(-i, st.mapmove_y) == 12:
                                st.mapmove_x, st.mapmove_y = -7, 0
                                break
                        else:
                            for j in xrange(1, 1 if st.mapmove_y else 8):
                                if colorat(st.mapmove_x, j) == 12:
                                    st.mapmove_x, st.mapmove_y = 0, 7
                                    break
                                if colorat(st.mapmove_x, -j) == 12:
                                    st.mapmove_x, st.mapmove_y = 0, -7
                                    break
                            else:
                                st.mapmove_x, st.mapmove_y = 0, 0
                    else:
                        st.mapmove_x, st.mapmove_y = 0, 0
                if colorat(st.mapmove_x, st.mapmove_y) != 12:
                    st.mapmove_x, st.mapmove_y = 0, 0
            mgl[4] += st.mapmove_x
            mgl[5] += st.mapmove_y
            if color != 10:
                if color != 12:
                    mgl[0] = color
                    self.clearstate()
                elif not st.walk_index and mgl[0] != 2 and not st.inputshift and randint(0, 99) < mgl[6]:
                    mgl[0] = 3
                    self.clearstate()
            img = withmask(st.loaded_frames[mgl[1]].offsets[0].frame.value)

            dx = mgl[4] - st.drawing_w / 2 # center the view over the character
            dy = mgl[5] - st.drawing_h / 2
            # TODO make safe
            viewrect = wx.Rect(st.drawing_x, st.drawing_y, st.drawing_w, st.drawing_h)
            self.frameregion.UnionRect(viewrect)
            x, y, w, h = viewrect
            imgto = self.framebuffer[y:y+h, x:x+w]
            imgfr = img.nparray[dy:dy+h, dx:dx+w]
            imgto[...] = imgfr[...]

            if st.walk_sprite_index == 0:
                st.walk_sprite_index = 1
            elif st.walk_sprite_index == 1:
                st.walk_sprite_index = 0
            elif st.walk_sprite_index == 2:
                st.walk_sprite_index = 1
            elif st.walk_sprite_index == 3:
                st.walk_sprite_index = 2
            st.move_trail[st.trail_index][:] = mgl[4], mgl[5], 3 * st.mapmove_orient + st.walk_sprite_index
            characters = []
            for i in xrange(mgl[7]):
                j = (st.trail_index - 6 * i) % 64
                characters.append((
                    st.move_trail[j][0] - st.move_trail[st.trail_index][0],
                    st.move_trail[j][1] - st.move_trail[st.trail_index][1],
                    12 * mgl[8+i] + st.move_trail[j][2]))
            if st.mapmove_orient != 1:
                characters.reverse()
            characters.sort(key=lambda e: e[1])
            for x, y, i in characters:
                self.drawcharacter(x, y, i, mgl[3])
            if st.mapmove_x or st.mapmove_y:
                st.trail_index = (st.trail_index + 1) % 64

        if mgl[12]:
            st.fadein(mgl[13], (mgl[14], mgl[14], mgl[14]))
            mgl[12] = 0
            if mgl[0] == 2:
                st.state = 4
        elif mgl[0] == 2:
            self.clearstate()
        mgl[15] = st.mapmove_orient

    def drawcharacter(self, x, y, j, i):
        st = self.states
        img = withmask(st.loaded_frames[i].offsets[j].frame.value)
        # TODO make safe
        w, h, = img.w, img.h
        viewrect = wx.Rect(
            st.drawing_x + st.drawing_w / 2 + x - w / 2,
            st.drawing_y + st.drawing_h / 2 + y - h / 2,
            w, h)
        x, y, w, h = viewrect
        imgto = self.framebuffer[y:y+h, x:x+w]
        x, y = 0, 0
        imgfr = img.nparray[y:y+h, x:x+w]
        mask = img.masks[0x0C][y:y+h, x:x+w]
        imgto[mask] = imgfr[mask]

    # dialogs

    def askforname(self):
        wx.CallAfter(self.shownamedialog)

    def shownamedialog(self):
        with wx.TextEntryDialog(self,  '', 'Name Input', DEFAULT_HERO) as dlg:
            dlg.CenterOnParent()
            dlg.ShowModal()
            self.states.hero_name = dlg.GetValue()
        self.clearstate()

    def showdialoguepopup(self, show=False, update=False):
        self.dialoguepopup.Show(show)
        if show and update:
            self.paintdialoguepopup()

    def paintdialoguepopup(self, dc=None):
        st = self.states
        if dc is None:
            dc = wx.ClientDC(self.dialoguepopup)
        if st.dlgpop_bmp is None:
            dc.SetPen(wx.BLACK_PEN)
            dc.SetBrush(wx.BLACK_BRUSH)
            dc.DrawRectangle(0, 0, st.dlgpop_w, st.dlgpop_h)
        else:
            dc.DrawBitmap(wx.BitmapFromImage(wx.ImageFromStream(StringIO(st.dlgpop_bmp))), 0, 0)
        self.SetFocus()

    def toggledialoguepopup(self):
        states = self.states
        show = not states.dlgpop_shown
        self.dialoguepopup.Show(show)
        states.dlgpop_shown = show

    def getdialoguetarget(self, lt=True):
        aw = self.states.use_dlgpop
        if aw == 2:
            return self.dialoguepopup
        elif aw <= 2 if aw else aw != 2:
            return self.bmppanel

SINES = np.array([
       0,    25,    49,    74,    98,   122,   147,   171,   195,
     219,   243,   267,   290,   314,   337,   360,   383,   405,
     428,   450,   471,   493,   514,   535,   556,   576,   596,
     615,   634,   653,   672,   690,   707,   724,   741,   757,
     773,   788,   803,   818,   831,   845,   858,   870,   882,
     893,   904,   914,   924,   933,   942,   950,   957,   964,
     970,   976,   981,   985,   989,   992,   995,   997,   999,
    1000,  1000,  1000,   999,   997,   995,   992,   989,   985,
     981,   976,   970,   964,   957,   950,   942,   933,   924,
     914,   904,   893,   882,   870,   858,   845,   831,   818,
     803,   788,   773,   757,   741,   724,   707,   690,   672,
     653,   634,   615,   596,   576,   556,   535,   514,   493,
     471,   450,   428,   405,   383,   360,   337,   314,   290,
     267,   243,   219,   195,   171,   147,   122,    98,    74,
      49,    25,     0,   -25,   -49,   -74,   -98,  -122,  -147,
    -171,  -195,  -219,  -243,  -267,  -290,  -314,  -337,  -360,
    -383,  -405,  -428,  -450,  -471,  -493,  -514,  -535,  -556,
    -576,  -596,  -615,  -634,  -653,  -672,  -690,  -707,  -724,
    -741,  -757,  -773,  -788,  -803,  -818,  -831,  -845,  -858,
    -870,  -882,  -893,  -904,  -914,  -924,  -933,  -942,  -950,
    -957,  -964,  -970,  -976,  -981,  -985,  -989,  -992,  -995,
    -997,  -999, -1000, -1000, -1000,  -999,  -997,  -995,  -992,
    -989,  -985,  -981,  -976,  -970,  -964,  -957,  -950,  -942,
    -933,  -924,  -914,  -904,  -893,  -882,  -870,  -858,  -845,
    -831,  -818,  -803,  -788,  -773,  -757,  -741,  -724,  -707,
    -690,  -672,  -653,  -634,  -615,  -596,  -576,  -556,  -535,
    -514,  -493,  -471,  -450,  -428,  -405,  -383,  -360,  -337,
    -314,  -290,  -267,  -243,  -219,  -195,  -171,  -147,  -122,
     -98,   -74,   -49,   -25])
COSINES = np.array([
    1000,  1000,   999,   997,   995,   992,   989,   985,   981,
     976,   970,   964,   957,   950,   942,   933,   924,   914,
     904,   893,   882,   870,   858,   845,   831,   818,   803,
     788,   773,   757,   741,   724,   707,   690,   672,   653,
     634,   615,   596,   576,   556,   535,   514,   493,   471,
     450,   428,   405,   383,   360,   337,   314,   290,   267,
     243,   219,   195,   171,   147,   122,    98,    74,    49,
      25,     0,   -25,   -49,   -74,   -98,  -122,  -147,  -171,
    -195,  -219,  -243,  -267,  -290,  -314,  -337,  -360,  -383,
    -405,  -428,  -450,  -471,  -493,  -514,  -535,  -556,  -576,
    -596,  -615,  -634,  -653,  -672,  -690,  -707,  -724,  -741,
    -757,  -773,  -788,  -803,  -818,  -831,  -845,  -858,  -870,
    -882,  -893,  -904,  -914,  -924,  -933,  -942,  -950,  -957,
    -964,  -970,  -976,  -981,  -985,  -989,  -992,  -995,  -997,
    -999, -1000, -1000, -1000,  -999,  -997,  -995,  -992,  -989,
    -985,  -981,  -976,  -970,  -964,  -957,  -950,  -942,  -933,
    -924,  -914,  -904,  -893,  -882,  -870,  -858,  -845,  -831,
    -818,  -803,  -788,  -773,  -757,  -741,  -724,  -707,  -690,
    -672,  -653,  -634,  -615,  -596,  -576,  -556,  -535,  -514,
    -493,  -471,  -450,  -428,  -405,  -383,  -360,  -337,  -314,
    -290,  -267,  -243,  -219,  -195,  -171,  -147,  -122,   -98,
     -74,   -49,   -25,     0,    25,    49,    74,    98,   122,
     147,   171,   195,   219,   243,   267,   290,   314,   337,
     360,   383,   405,   428,   450,   471,   493,   514,   535,
     556,   576,   596,   615,   634,   653,   672,   690,   707,
     724,   741,   757,   773,   788,   803,   818,   831,   845,
     858,   870,   882,   893,   904,   914,   924,   933,   942,
     950,   957,   964,   970,   976,   981,   985,   989,   992,
     995,   997,   999,  1000])

MAP_KEY_XYO = {
    0b0010: ( 0,  7, 0),
    0b1010: ( 5,  5, 0),
    0b1000: ( 7,  0, 3),
    0b1001: ( 5, -5, 1),
    0b0001: ( 0, -7, 1),
    0b0101: (-5, -5, 1),
    0b0100: (-7,  0, 2),
    0b0110: (-5,  5, 0),
}

if __name__ == '__main__':
    try:
        __file__
    except NameError:
        __file__ = sys.executable
    if len(sys.argv) > 1:
        sgspath = sys.argv[1]
    else:
        sgspath = os.path.abspath('SGS.DAT')
        if not os.path.exists(sgspath):
            print >>sys.stderr, 'SGS.DAT not found in current directory.'
            print >>sys.stderr, 'Run from game directory or supply path to .DAT file on the command line.'

    paud = pyaudio.PyAudio()
    app = wx.PySimpleApp()
    frame = GameFrame(None)
    app.SetTopWindow(frame)
    frame.app = app
    frame.Show(True)
    frame.SetDatPath(sgspath)

    if len(sys.argv) > 3:
        frame.SetWinName(sys.argv[2]+'.WIN')
        frame.SetStartAddress(int(sys.argv[3], 16))
        n = 5
        while len(sys.argv) > n:
            frame.states.globals[int(sys.argv[n-1])] = int(sys.argv[n])
            n += 2

    frame.StartGame()
    app.MainLoop()
    paud.terminate()
#
