# encoding=utf_8
from construct import *
import re

# 0x814F \uFF3E ＾
# 0x818F \uFFE5 ￥
# 0x8190 \uFF04 ＄ Replace with player name (default: 中田)

BREAK_PATTERN = re.compile(u'(?:\1(.))?([^\1￥]*)(￥)?', re.S)

def breaktext(s):
    for m in BREAK_PATTERN.finditer(s):
        yield (ord(m.group(1)) if m.group(1) else None, m.group(2), bool(m.group(3)))

def jointext(lines):
    parts = []
    for n, s, b in lines:
        if n is not None:
            parts.append(u'\1%s'%chr(n))
        if s:
            parts.append(s)
        if b:
            parts.append(u'￥')
    return ''.join(parts)

# 01h VK_NUMPAD8    VK_UP
# 02h VK_NUMPAD2    VK_DOWN  
# 04h VK_NUMPAD4    VK_LEFT
# 08h VK_NUMPAD6    VK_RIGHT
# 10h VK_NUMPAD0    VK_ESCAPE   WM_RBUTTONDOWN  
# 20h VK_F11        VK_F12      VK_SPACE    WM_LBUTTONDOWN  VK_RETURN   VK_CONTROL

def EStruct(name, *args, **kwargs):
    return Struct(None, Value('fourcc', lambda x: name), *args, **kwargs)

Instruction = Struct('bytecode',
    Anchor('address'),
    Byte('instruction'),
    #Probe(),
    Embed(Switch('args', lambda ctx: ctx.instruction, {
        0x00: EStruct('exit'),
        0x01: EStruct('offset', ULInt16('x'), ULInt16('y')),
        0x02: EStruct('dims', ULInt16('w'), ULInt16('h')),
        0x03: EStruct('anmload', Byte('a'), CString('path', encoding='sjis')),
        0x04: EStruct('anmfree', Byte('a'), ),
        0x05: EStruct('fadein', Byte('a'), SLInt16('div')),
        0x06: EStruct('fadeout', Byte('a'), SLInt16('div')),
        0x07: EStruct('anmplt', Byte('a')),
        0x08: EStruct('pausefade'),
        0x09: EStruct('stopfade'),
        0x0A: EStruct('frameseq',
            Byte('a'),
            ULInt16('b'),
            ULInt16('c'),
            ULInt32('d'),
            OptionalGreedyRange(Struct('e',
                NoneOf(Byte('frame'), [0]),
                Byte('b'),
            )),
            Const(Byte(None), 0),
        ),
        0x0B: EStruct('rmlayer', Byte('level')),
        0x0C: EStruct('waitforlayer', Byte('level')),
        0x0D: EStruct('rmall'),
        0x0E: EStruct('nextwait'),
        0x0F: EStruct('nextwait'), # is 0x0E
        0x10: EStruct('prepanm', Byte('a')),
        0x12: EStruct('copyimg', Byte('a'), Byte('b'), Byte('c'), ULInt16('d'), ULInt16('e')),
        0x13: EStruct('delaywait', ULInt16('a')), # sleeps/waits for <a> 1/100 of a second (100 = 1 sec = 24 frames)
        0x14: EStruct('jmp', ULInt16('target')), # jmp
        0x15: EStruct('continuehandler', ULInt16('target')),
        0x16: EStruct('escapehandler', ULInt16('target')),
        0x17: EStruct('tweenmove', Byte('a'), ULInt16('b'), ULInt16('c'), ULInt16('d'), ULInt16('e'), ULInt16('f'), ULInt16('g'), ULInt16('h')),
        0x18: EStruct('motionwait', Byte('a')),
        0x19: EStruct('bgmplay', Byte('a'), Byte('sndnum')),
        0x1A: EStruct('bgmstop'),
        0x1B: EStruct('definebgm', Byte('sndnum'), CString('path', encoding='sjis')),
        0x1D: EStruct('pcmplay', Byte('flags'), CString('path', encoding='sjis')),
        0x1E: EStruct('pcmstop'),
        0x1F: EStruct('pcmwait'),
        0x20: EStruct('textarea', ULInt16('x'), ULInt16('y'), Byte('w'), Byte('h')), # x and y are in pixels, w in halfwidth characters, and h in lines
        0x21: EStruct('print', CString('text', encoding='sjis')), # prints text in the textarea
        0x22: EStruct('wrappad', Byte('a')),
        0x23: EStruct('fontsize', Byte('a')), # MS Gothic
        0x24: EStruct('string', Byte('name'), CString('text', encoding='sjis')),
        0x25: EStruct('call', ULInt16('target')),
        0x26: EStruct('return'),
        0x27: EStruct('fullscreen', Byte('a')), # turns on and off fullscreen
        0x28: EStruct('pltxform', # dissolve to altered palette
            Byte('type'), 
            ULInt16('frames'), 
            Byte('n'), 
            MetaRepeater(lambda ctx: ctx.n, Struct('colors',
                Byte('i'),
                Byte('r'),
                Byte('g'),
                Byte('b'),
            )),
        ),
        0x29: EStruct('exithandler', ULInt16('target')),
        0x2A: EStruct('disableexit'),
        0x2B: EStruct('loadhandler', ULInt16('target')),
        0x2C: EStruct('disableload'),
        0x2D: EStruct('disableskip'),
        0x2E: EStruct('je', Byte('a'), ULInt16('b'), ULInt16('target')), # jump if equal
        0x2F: EStruct('jne', Byte('a'), ULInt16('b'), ULInt16('target')), # jump if not equal
        0x30: EStruct('jle', Byte('a'), ULInt16('b'), ULInt16('target')), # jump if b >= [a]
        0x31: EStruct('jge', Byte('a'), ULInt16('b'), ULInt16('target')), # jump if b <= [a]
        0x32: EStruct('switch', Byte('a')), # store a for switch?
        0x33: EStruct('caseeq', ULInt16('a'), ULInt16('target')), # jump if equal to stored
        0x34: EStruct('setv', Byte('a'), ULInt16('b')), # *a = b
        0x35: EStruct('incv', Byte('a')), # ++(*a)
        0x36: EStruct('decv', Byte('a')), # --(*a)
        0x37: EStruct('addv', Byte('a'), ULInt16('b')), # *a += b
        0x38: EStruct('orv', Byte('a'), ULInt16('b')), # *a |= b
        0x39: EStruct('andv', Byte('a'), ULInt16('b')), # *a &= b
        0x3A: EStruct('blackout', ULInt16('a')),
        0x3B: EStruct('loadsystem'), # load SaveSystem
        0x3C: EStruct('savesystem'), # save SaveSystem
        0x3D: EStruct('cleararea', ULInt16('x'), ULInt16('y'), ULInt16('w'), ULInt16('h')), # clear rectangle
        0x3E: EStruct('choice',
            Byte('a'),
            ULInt16('b'),
            ULInt16('c'),
            Byte('d'),
            Byte('n'),
            MetaRepeater(lambda ctx: ctx.n, Struct('f',
                CString('text', encoding='sjis')
            )),
        ),
        0x3F: EStruct('choicelayout', Byte('a')),
        0x40: EStruct('enablesave', Byte('a')),
        0x41: EStruct('pltset',
            Byte('n'),
            MetaRepeater(lambda ctx: ctx.n, Struct('b',
                Byte('i'),
                Byte('r'),
                Byte('g'),
                Byte('b'),
            )),
        ),
        0x42: EStruct('readwait'),
        0x43: EStruct('readfactor', ULInt16('a')),
        0x44: EStruct('inputwait', Byte('a')), # later, *a = inputflags
        0x45: EStruct('storeinput', Byte('a')), # *a = inputflags
        0x46: EStruct('noinputwait'),
        0x47: EStruct('minorreset', Byte('a')), # doesn't actually use a. Resets the animation layers, frames, the stack, 
        0x48: EStruct('caseand', ULInt16('a'), ULInt16('target')), # jump if !(*a & switchval)
        0x49: EStruct('mouseover', Byte('a'), Byte('b'), Byte('storeat')),
        0x4A: EStruct('loadwin', CString('path', encoding='sjis')), # load .win and jump to start
        0x4B: EStruct('caretpos', ULInt16('x'), ULInt16('y')), # the position of the blinking continue indicator
        0x4C: EStruct('setscroll', ULInt16('a'), ULInt16('b'), ULInt16('c'), ULInt16('d')),
        0x4D: EStruct('resumeanim'),
        0x4E: EStruct('pauseanim'),
        0x4F: EStruct('pauseafter', Byte('level')),
        0x50: EStruct('fadein2', Byte('r'), Byte('g'), Byte('b'), SLInt16('div')), # fade in from a color
        0x51: EStruct('fadeout2', Byte('r'), Byte('g'), Byte('b'), SLInt16('div')), # fade out to a color
        0x52: EStruct('shimmer', Byte('a')),
        0x53: EStruct('repeathandler', ULInt16('target')),
        0x54: EStruct('disablerepeat'),
        0x55: EStruct('colorrange', Byte('a'), Byte('b')),
        0x56: EStruct('loadsave', CString('a')),
        0x57: EStruct('savesave', CString('a')),
        0x58: EStruct('movv', Byte('a'), Byte('b')), # *a = *b
        0x59: EStruct('wave', Byte('a'), ULInt16('b'), ULInt16('c'), ULInt16('d'), ULInt16('e'), ULInt16('f')),
        0x5A: EStruct('wavewait', Byte('a')),
        0x5B: EStruct('pointmove',
            Byte('a'),
            Byte('n'),
            MetaRepeater(lambda ctx: ctx.n, Struct('c',
                ULInt16('a'),
                ULInt16('b'),
            )),
        ),
        0x5C: EStruct('stoptween', Byte('level')),
        0x5D: EStruct('bgmvolume', Byte('volume')), # set bgm volume to volume/255 of user's bgm volume
        0x5E: EStruct('bgmfade', ULInt16('frames')), # fade out bgm to silence over n frames
        0x5F: EStruct('nop'),
        0x60: EStruct('bgmfade2', Byte('volume'), ULInt16('frames')), # set bgm volume to volume/255 of user's bgm volume over n frames
        0x61: EStruct('newimg', Byte('a'), ULInt16('x'), ULInt16('y'), ULInt16('w'), ULInt16('h')),
        0x62: EStruct('enablepopup', Byte('a')),
        0x63: EStruct('disablesfx', Byte('flags')), # prevents voice and sfx with flags in common from playing. 
        0x64: EStruct('pltshift',
            Byte('n'),
            MetaRepeater(lambda ctx: ctx.n, Struct('a',
                ULInt16('frames'),
                Byte('n'),
                MetaRepeater(lambda ctx: ctx.n, Struct('b',
					Byte('i'),
					Byte('r'),
					Byte('g'),
					Byte('b'),
				)),
            )),
        ),
        0x65: EStruct('tracks', Byte('a')),
        0x66: EStruct('mousemoved', Byte('a')), # forces mouseover to get the color without the mouse being moved
        0x67: EStruct('enablei18n'),
        0x68: EStruct('printi18n', ULInt16('line')),
        0x69: EStruct('choicei18n',
            Byte('a'),
            ULInt16('b'),
            ULInt16('c'),
            Byte('d'),
            MetaRepeater(lambda ctx: ctx.d, Struct('e', ULInt16('a'))),
        ),
        0x6A: EStruct('rand', Byte('a'), ULInt16('mod')), # *a = rand() % mod
        0x6B: EStruct('playsfx', Byte('flags'), ULInt16('delay'), CString('path', encoding='sjis')),
        0x6C: EStruct('looppcm', Byte('a')),
        0x6D: EStruct('jmplayer', Byte('a'), ULInt16('target')), # jump if layer a is not deleted
        0x6E: EStruct('getkey', Byte('a'), Byte('b')), # Eventually, *a = 
        0x6F: EStruct('je2', Byte('a'), Byte('b'), ULInt16('target')), # jump if [a] == [b]
        0x70: EStruct('jne2', Byte('a'), Byte('b'), ULInt16('target')), # jump if [a] != [b]
        0x71: EStruct('jle2', Byte('a'), Byte('b'), ULInt16('target')), # jump if [a] <= [b]
        0x72: EStruct('jge2', Byte('a'), Byte('b'), ULInt16('target')), # jump if [a] >= [b]
        0x73: EStruct('textdraw', CString('a', encoding='sjis')),
        0x74: EStruct('textcolor', Byte('r'), Byte('g'), Byte('b')),
        0x75: EStruct('textpos', ULInt16('x'), ULInt16('y')),
        0x76: EStruct('heroname'),
        0x78: EStruct('playout', Byte('disable')), # disables clicking to end a pcm
        0x79: EStruct('mapvar', Byte('a')),
        0x7A: EStruct('casenand', ULInt16('a'), ULInt16('target')), # jump if *a & switchval
        0x7B: EStruct('popupbg', ULInt16('a'), ULInt16('b'), CString('path', encoding='sjis')),
    })),
    Anchor('nextaddr'),
)

bytecode = GreedyRange(Instruction)

if __name__ == '__main__':
    import sys
    
    with open(sys.argv[1], 'rb') as f:
        data = f.read()

    print bytecode.parse(data)